// calc.c
// voltage independent conductance
// 03,12,09, Jin

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "board.h"
#include "calc.h"
#include "com.h"
#include "global.h"
#include "cycle_counter.h"

#define PI		3.141592654

#define IA		16807
#define IM		2147483647
#define AM		(1.0/IM)
#define IQ		127773
#define IR		2836
#define NTAB	32
#define NDIV	(1+(IM-1)/NTAB)
#define EPS		1.2e-7
#define RNMX	(1.0-EPS)

volatile double	*lookupTable	= (void *) LOOKUP_TABLE;

/****************************************************************************************/
//
double ran1( long *idum )
{
	int			j;
	long		k;
	static long	iy = 0;
	static long	iv[NTAB];
	double		temp;

	if (*idum <= 0 || !iy)
	{
		if (-(*idum) < 1) *idum = 1;
		else *idum = -(*idum);
		for (j = NTAB + 7; j >= 0; j--)
		{
			k = (*idum) / IQ;
			*idum = IA * (*idum - k * IQ) - IR * k;
			if (*idum < 0) *idum += IM;
			if (j < NTAB) iv[j] = *idum;
		}
		iy = iv[0];
	}
	k = (*idum)/IQ;
	*idum = IA * (*idum - k * IQ) - IR * k;
	if (*idum < 0) *idum += IM;
	j = iy/NDIV;
	iy = iv[j];
	iv[j] = *idum;
	if ((temp = AM * iy) > RNMX) return RNMX;
	else return temp;
}


/****************************************************************************************/

int	FindLastD( int n, double *t0, double t )
{
	int		i;

	for (i = 0; i < n; i++)
	{
		if (t0[i] >= t) break;
	}
//	if (i >= n) return -1;
	return (i - 1);
}



/****************************************************************************************/
//
double gammln( double xx )
//Returns the value ln[Γ(xx)] for xx > 0.
{
	// Internal arithmetic will be done in double precision, a nicety that you can omit if five-figure
	// accuracy is good enough.
	double			x, y, tmp, ser;
	static double	cof[6] =	{
									 76.18009172947146,
									-86.50532032941677,
									 24.01409824083091,
									-1.231739572450155,
									 0.1208650973866179e-2,
									-0.5395239384953e-5
								};
	int 			j;

	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++) ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}
/****************************************************************************************/
//
double binorand( int n, double pp, long *idum )
//Returns as a floating-point number an integer value that is a random deviate drawn from
//a binomial distribution of n trials each of probability pp, using ran1(idum) as a source of
//uniform random deviates.
{

	int j;
	static int nold=(-1);
	double am,em,g,angle,p,bnl,sq,t,y;
	static double pold=(-1.0),pc,plog,pclog,en,oldg;

	p=(pp <= 0.5 ? pp : 1.0-pp);
	//The binomial distribution is invariant under changing pp to 1-pp, if we also change the
	//answer to n minus itself; we’ll remember to do this below.
	am=n*p; 				//This is the mean of the deviate to be produced.
	if (n < 25)				//Use the direct method while n is not too large.
	{
		bnl=0.0;			//This can require up to 25 calls to ran1.
		for (j=1;j<=n;j++)
		{
			if (ran1(idum) < p) ++bnl;
//			if ( (double)rand()/(double)(RAND_MAX) < p) ++bnl;
		}
	}
	else if (am < 1.0)		//If fewer than one event is expected out of 25
	{
							//or more trials, then the distribution is quite
							//accurately Poisson. Use direct Poisson method.
		g=exp(-am);
		t=1.0;
		for (j=0;j<=n;j++)
		{
			t *= ran1(idum);
//			t *= (double)rand()/(double)(RAND_MAX);
			if (t < g) break;
		}
		bnl=(j <= n ? j : n);
	}
	else					//Use the rejection method.
	{
		if (n != nold)		//If n has changed, then compute useful quantities
		{
			en=n;
			oldg=gammln(en+1.0);
			nold=n;
		}

		if (p != pold)		//If p has changed, then compute useful quantities
		{
			pc=1.0-p;
			plog=log(p);
			pclog=log(pc);
			pold=p;
		}
		sq=sqrt(2.0*am*pc); //The following code should by now seem familiar:
							 //rejection method with a Lorentzian comparison
							 //function.
		do
		{
			do
			{
				angle=PI*ran1(idum);
//				angle=PI*(double)rand()/(double)(RAND_MAX);
				y=tan(angle);
				em=sq*y+am;
			} while (em < 0.0 || em >= (en+1.0)); //Reject.
			em=floor(em); //Trick for integer-valued distribution.
			t=1.2*sq*(1.0+y*y)*exp(oldg-gammln(em+1.0)
									   -gammln(en-em+1.0)+em*plog+(en-em)*pclog);
		} while (ran1(idum) > t);
//		} while ((double)rand()/(double)(RAND_MAX+1.0) > t); //Reject. This happens about 1.5 times per devi
		bnl=em;                   //ate, on average.
	}
	if (p != pp) bnl=n-bnl; //Remember to undo the symmetry transformation
	return bnl;
}

/*************************************************************************************/
//
double normrand( long *idum ) // from numerical recipes
{
	//Returns a normally distributed deviate with zero mean and unit variance, using ran1(idum)
	//as the source of uniform deviates.
	static int iset=0;
	static double gset;
	double fac,rsq,v1,v2;

	if (*idum < 0) iset=0;
	if  (iset == 0)
	{
		do
		{
			v1=2.0*ran1(idum)-1.0;
			v2=2.0*ran1(idum)-1.0;
	//		v1=2.0*(double)rand()/(double)(RAND_MAX) - 1.0; //pick two uniform numbers in the square ex
	//		v2=2.0*(double)rand()/(double)(RAND_MAX) - 1.0; //tending from -1 to +1 in each direction,
			rsq=v1*v1+v2*v2; //see if they are in the unit circle,

		} while (rsq >= 1.0 || rsq == 0.0); //and if they are not, try again.
		fac=sqrt(-2.0*log(rsq)/rsq);
		//Now make the Box-Muller transformation to get two normal deviates. Return one and
		//save the other for next time.
		gset=v1*fac;
		//iset=1; //Set flag.
		return v2*fac;

	}
	else
	{ //We have an extra deviate handy,
		iset=0; //so unset the flag,
		return gset; //and return it.
	}

}
/********************************************************************************************/
//
void CalcDynamicG(double *t0, int nt, int nr, double pr, double sigma, double taur, double tauf, double g_unit)
{
	double *t_avail=NULL;
	int n_avail=nr;
	double n_release=0.0;
	double uniranm;
	double p=0.0;
	double p0=0.0;

	int i, j,k,counter;
	char *index = NULL;

	long	idum = (Get_sys_count()) & 0x7FFFFFFF;

//	srand(rands);

	if( (t_avail = malloc(sizeof(double) * nr) ) == NULL )
	{
		return;

	}


	for (i=0;i<nr;i++)
	{
		t_avail[i]=t0[0];
	}

	for (i=0;i<(nt-1);i++)  //
	{
		if( (index = malloc(sizeof(char) * nr) ) == NULL )
		{
			return;

		}

		p=p0 + pr*(1-p0);

		// find
		counter=0;
		for (j=0;j<nr;j++)
		{
			if(t0[i] >= t_avail[j])
			{
				index[j] = 1;
				counter++;
			}
			else
			{
				index[j] = 0;
			}
		}
		///////
		if (counter==0)
		{
			n_avail=0;
			gDGMax[i]=0.0;
		}
		else
		{
			n_avail=counter;
			n_release=binorand(n_avail, p, &idum);

			if (n_release==0.0)
			{
				gDGMax[i]=0.0;
			}
			else
			{
				do
       				gDGMax[i]= (normrand( &idum ) * (double)sqrt(n_release * pow(sigma,2.0)) + n_release) * g_unit;
				while (gDGMax[i]<0.0 || gDGMax[i]>2.0*n_release*g_unit);

				for (j=0,k=0;j<nr && k<=(int)n_release;j++)
				{
					if (index[j] == 1)
					{
						do
						{
							uniranm = ran1( &idum );
//							uniranm = (double)rand()/(double)(RAND_MAX+1.0);
						} while(uniranm==0);

						t_avail[j] = t0[i] + (-1.0)*taur*log(uniranm);
						k++;
					}
				}
			}
		}
		free(index);

		p0=p*exp((-1)*(t0[i+1]-t0[i])/tauf);
	}


		// calculate the last one
		if( (index = malloc(sizeof(char) * nr) ) == NULL )
		{
			return;

		}

		p=p0 + pr*(1-p0);

		// find
		counter=0;
		for (j=0;j<nr;j++)
		{
			if(t0[i] >= t_avail[j])
			{
				index[j] = 1;
				counter++;
			}
			else
			{
				index[j] = 0;
			}
		}
		///////
		if (counter==0)
		{
			n_avail=0;
			gDGMax[i]=0.0;
		}
		else
		{
			n_avail=counter;
			n_release=binorand(n_avail, p, &idum);

			if (n_release==0.0)
			{
				gDGMax[i]=0.0;
			}
			else
			{
				do
       				gDGMax[i]= (normrand( &idum ) * (double)sqrt(n_release * pow(sigma,2.0)) + n_release) * g_unit;
				while (gDGMax[i]<0.0 || gDGMax[i]>2.0*n_release*g_unit);

				for (j=0,k=0;j<nr && k<=(int)n_release;j++)
				{
					if (index[j] == 1)
					{
						do
						{
							uniranm = ran1( &idum );
//							uniranm = (double)rand()/(double)(RAND_MAX+1.0);
						} while(uniranm==0);

						t_avail[j] = t0[i] + (-1.0)*taur*log(uniranm);
						k++;
					}
				}
			}
		}
		free(index);

	free(t_avail);

}
/**********************************************************************************************/
//
void MakeGabaAlphaLT( double *t0, int n, int numVal, double period, int gain, double g_max, double t_peak )
{
	double	g_const;
	double	t = 0.0;

	double	delay_t, result;
	long	i, j, index;


	g_const = g_max * exp(1) / t_peak;
	t = t0[0];

	for (i = 0; i < numVal; i++)
	{
		result = 0.0;

		index = FindLastD( n, t0, t );
		if (index >= 0)
		{
			for (j = 0; j <= index; j++)
			{
				delay_t = t - t0[j];
				result = result + (-1.0 * g_const * delay_t * exp(-delay_t/t_peak));
			}
		}

		switch (gain)
		{
			case 0:
				result /= 1.0e-10;
				break;
			case 1:
				result /= 1.0e-9;
				break;
			case 2:
				result /= 1.0e-8;
				break;
			case 3:
				result /= 1.0e-7;
				break;
		}
		t += period;
		lookupTable[i] = result;
	}
}


/****************************************************************************************/

void MakeDynamicAlphaLT(	double	*t0,
							int		n,
							int		numVal,
							double	period,
							int		gain,
							int		d_nr,
							double	d_pr,
							double	d_sigma,
							double	d_tau,
							double  d_tauf,
							double	d_gunit,
							double	t_peak
						)
{
	double	g_const;
	double	t = 0.0;

	double	delay_t, result;
	long	i, j, index;

   // g_max[0]=1;
    //g_max[1]=2;

	t = t0[0];

	CalcDynamicG( t0, n, d_nr, d_pr,d_sigma,d_tau, d_tauf, d_gunit );

	for (i = 0; i < numVal; i++)
	{

		result = 0.0;

		index = FindLastD( n, t0, t );
		if (index >= 0)
		{
			for (j = 0; j <= index; j++)
			{
				delay_t = t - t0[j];
				g_const = gDGMax[j] * exp(1) / t_peak;
				result = result + (-1.0 * g_const * delay_t * exp(-delay_t/t_peak));
			}
		}

		switch (gain)
		{
			case 0:
				result /= 1.0e-10;
				break;
			case 1:
				result /= 1.0e-9;
				break;
			case 2:
				result /= 1.0e-8;
				break;
			case 3:
				result /= 1.0e-7;
				break;
		}
		t += period;
		lookupTable[i] = result;
	}
}


/****************************************************************************************/

void MakeGaba2ExpLT( double *t0, int n, int numVal, double period, int gain, double gConst, double tauR, double tauD )
{
	double	t = 0.0;

	double	delay_t, result, scaler;
	long	i, j, index;


	//g_const = g_max * exp(1) / t_peak;
	t = t0[0];

	scaler = 1.0 / (exp( log(tauR/tauD) * tauR/(tauD-tauR)) - exp( log(tauR/tauD) * tauD/(tauD-tauR) ));

	for (i = 0; i < numVal; i++)
	{
		result = 0.0;

		index = FindLastD( n, t0, t );
		if (index >= 0)
		{
			for (j = 0; j <= index; j++)
			{
				delay_t = t - t0[j];
				result = result + (-1.0 * gConst * scaler *( 1.0 - exp(-delay_t/tauR)) * exp(-delay_t/tauD));
			}
		}

		switch (gain)
		{
			case 0:
				result /= 1.0e-10;
				break;
			case 1:
				result /= 1.0e-9;
				break;
			case 2:
				result /= 1.0e-8;
				break;
			case 3:
				result /= 1.0e-7;
				break;
		}
		t += period;
		lookupTable[i] = result;
	}
}


/****************************************************************************************/

void MakeAmpaLT(	double	*t0,
					int		n,
					int		numVal,
					double	period,
					int		gain,
					int		n_syn,
					double	g_max,
					double	g_sd,
					double	tauR,
					double	tauD
				)
{
	double	t = 0.0;

	double	delay_t, result, t_peak, conv;
	long	i, index, count = 0L;

	double g_max_rand;
	//double *g_const;
	long	idum = (Get_sys_count()) & 0x7FFFFFFF;

	double	ranNum;


	t_peak=tauR*tauD/(tauR-tauD)*log(tauR/tauD);
	conv=exp( log(tauR/tauD) * tauR/(tauD-tauR)) - exp( log(tauR/tauD) * tauD/(tauD-tauR) )  ;


//	srand(rands);

//	Printf( "n_syn = %d; g_sd = %g; g_max = %g\n", n_syn, g_sd, g_max );
	for (i=0; i<n; i++)
	{
		do
		{
			ranNum = normrand( &idum );
//			g_max_rand = (normrand( &idum ) * sqrt((double)n_syn * pow(g_sd,2.0)) + (double)n_syn * g_max);
//			g_max_rand = (normrand( &idum ) * sqrt((double)n_syn * g_sd * g_sd) + (double)n_syn * g_max);
			g_max_rand = (ranNum * sqrt((double)n_syn * g_sd * g_sd) + (double)n_syn * g_max);

//			Printf( "ranNum = %g; g_max_rand = %g\n", ranNum, g_max_rand );
			count++;
			if (count > 100) return;
		}
		while(g_max_rand<=0.0);

		gDGMax[i] = g_max_rand / conv;

	}

	t = t0[0];

	for (i = 0; i < numVal; i++)
	{
		result = 0.0;

		index = FindLastD( n, t0, t );
		if (index >= 0)
		{

			delay_t = t-t0[index];

			if (delay_t < 10.0 * t_peak)
			{
				result = (-1.0) * gDGMax[index] * (exp(-delay_t/tauD) - exp(-delay_t/tauR));
			}
		}
			//for (j = 0; j <= index; j++)
			//{
				//delay_t = t - t0[j];
				//result = result + (-1.0 * gConst *( 1.0 - exp(-delay_t/tauR)) * exp(-delay_t/tauD));
		//	}
		//}



		switch (gain)
		{
			case 0:
				result /= 1.0e-10;
				break;
			case 1:
				result /= 1.0e-9;
				break;
			case 2:
				result /= 1.0e-8;
				break;
			case 3:
				result /= 1.0e-7;
				break;
		}
		t += period;
		lookupTable[i] = result;
	}
}



/****************************************************************************************/

void MakeAmpaPnCLT(	double	*t0,
					int		n,
					int		numVal,
					double	period,
					int		gain,
					int		n_syn,
					double	g_max,
					double	g_sd,
					double	tauR,
					double	tauD
				)
{
	double	t = 0.0;

	double	delay_t, result, t_peak, conv;
	long	i, index, count = 0L;

	double g_max_rand;
	//double *g_const;
	long	idum = (Get_sys_count()) & 0x7FFFFFFF;

	double	ranNum;


	t_peak=tauR*tauD/(tauR-tauD)*log(tauR/tauD);
	conv=exp( log(tauR/tauD) * tauR/(tauD-tauR)) - exp( log(tauR/tauD) * tauD/(tauD-tauR) )  ;


//	srand(rands);

//	Printf( "n_syn = %d; g_sd = %g; g_max = %g\n", n_syn, g_sd, g_max );
	for (i=0; i<n; i++)
	{
		do
		{
			ranNum = normrand( &idum );
//			g_max_rand = (normrand( &idum ) * sqrt((double)n_syn * pow(g_sd,2.0)) + (double)n_syn * g_max);
//			g_max_rand = (normrand( &idum ) * sqrt((double)n_syn * g_sd * g_sd) + (double)n_syn * g_max);
			g_max_rand = (ranNum * sqrt((double)n_syn * g_sd * g_sd) + (double)n_syn * g_max);

//			Printf( "ranNum = %g; g_max_rand = %g\n", ranNum, g_max_rand );
			count++;
			if (count > 100) return;
		}
		while(g_max_rand<=0.0);

		gDGMax[i] = g_max_rand / conv;

	}

	t = t0[0];

	for (i = 0; i < numVal; i++)
	{
		result = 0.0;

		index = FindLastD( n, t0, t );
		if (index >= 0)
		{

			delay_t = t-t0[index];

			if (delay_t < 10.0 * t_peak)
			{
				result = (-1.0) * gDGMax[index] * (exp(-delay_t/tauD) - exp(-delay_t/tauR));
			}
		}
			//for (j = 0; j <= index; j++)
			//{
				//delay_t = t - t0[j];
				//result = result + (-1.0 * gConst *( 1.0 - exp(-delay_t/tauR)) * exp(-delay_t/tauD));
		//	}
		//}



		switch (gain)
		{
			case 0:
				result /= 1.0e-10;
				break;
			case 1:
				result /= 1.0e-9;
				break;
			case 2:
				result /= 1.0e-8;
				break;
			case 3:
				result /= 1.0e-7;
				break;
		}
		t += period;
		lookupTable[i] = result;
	}
}



/****************************************************************************************/

double CalcI( unsigned long t, double vCell, double eRev, unsigned long first, unsigned long num )
{
	double			current = 0.0;
	unsigned long	timeIndex;

	if (t > first)
	{
		timeIndex = (t - first) % num;
		current = lookupTable[timeIndex] * (vCell - eRev);
	}

	return current;
}

